package repository

import (
	"context"
	"errors"
	"time"

	"code.30cm.net/digimon/app-cloudep-product-service/pkg/domain"
	"code.30cm.net/digimon/app-cloudep-product-service/pkg/domain/entity"
	"code.30cm.net/digimon/app-cloudep-product-service/pkg/domain/repository"
	mgo "code.30cm.net/digimon/library-go/mongo"
	"github.com/zeromicro/go-zero/core/stores/cache"
	"github.com/zeromicro/go-zero/core/stores/mon"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

type TagsRepositoryParam struct {
	Conf      *mgo.Conf
	CacheConf cache.CacheConf
	DBOpts    []mon.Option
	CacheOpts []cache.Option
}

type TagsRepository struct {
	Tags        mgo.DocumentDBWithCacheUseCase
	TageBinding mgo.DocumentDBWithCacheUseCase
}

func NewTagsRepository(param TagsRepositoryParam) repository.TagRepo {
	tags := entity.Tags{}
	tagsDB, err := mgo.MustDocumentDBWithCache(
		param.Conf,
		tags.CollectionName(),
		param.CacheConf,
		param.DBOpts,
		param.CacheOpts,
	)
	if err != nil {
		panic(err)
	}

	tagBinding := entity.TagsBindingTable{}
	bindingsDB, err := mgo.MustDocumentDBWithCache(
		param.Conf,
		tagBinding.CollectionName(),
		param.CacheConf,
		param.DBOpts,
		param.CacheOpts,
	)
	if err != nil {
		panic(err)
	}

	return &TagsRepository{
		Tags:        tagsDB,
		TageBinding: bindingsDB,
	}
}

func (repo *TagsRepository) Create(ctx context.Context, data *entity.Tags) error {
	if data.ID.IsZero() {
		now := time.Now().UTC().UnixNano()
		data.ID = primitive.NewObjectID()
		data.CreatedAt = now
		data.UpdatedAt = now
	}
	rk := domain.GetTagsRK(data.ID.Hex())
	_, err := repo.Tags.InsertOne(ctx, rk, data)

	return err
}

func (repo *TagsRepository) GetByID(ctx context.Context, id string) (*entity.Tags, error) {
	oid, err := primitive.ObjectIDFromHex(id)
	if err != nil {
		return nil, err
	}

	var result *entity.Tags
	err = repo.Tags.FindOne(ctx, domain.GetTagsRK(id), &result, bson.M{"_id": oid})
	switch {
	case err == nil:
		return result, nil
	case errors.Is(err, mon.ErrNotFound):
		return nil, ErrNotFound
	default:
		return nil, err
	}
}

func (repo *TagsRepository) Update(ctx context.Context, id string, data repository.TagModifyParams) error {
	now := time.Now().UTC().UnixNano()
	// 動態構建更新內容
	updateFields := bson.M{
		"updated_at": now, // 確保 `updateAt` 總是更新
	}
	if data.Name != nil {
		updateFields["name"] = *data.Name
	}
	if data.Types != nil {
		updateFields["types"] = *data.Types
	}
	if data.ShowType != nil {
		updateFields["show_type"] = *data.ShowType
	}
	if data.Cover != nil {
		updateFields["cover"] = *data.Cover
	}

	oid, err := primitive.ObjectIDFromHex(id)
	if err != nil {
		return ErrInvalidObjectID
	}

	// 執行更新
	rk := domain.GetTagsRK(id)
	_, err = repo.Tags.UpdateOne(ctx, rk, bson.M{"_id": oid}, bson.M{"$set": updateFields})
	if err != nil {
		return err
	}

	return err
}

func (repo *TagsRepository) Delete(ctx context.Context, id string) error {
	oid, err := primitive.ObjectIDFromHex(id)
	if err != nil {
		return ErrInvalidObjectID
	}

	filter := bson.M{"_id": oid}
	_, err = repo.Tags.DeleteOne(ctx, domain.GetTagsRK(id), filter)
	if err != nil {
		return err
	}

	return nil
}

func (repo *TagsRepository) List(ctx context.Context, params repository.TagQueryParams) ([]*entity.Tags, int64, error) {
	// 構建查詢過濾器
	filter := bson.M{}
	if params.Name != nil {
		filter["name"] = *params.Name
	}
	if params.Types != nil {
		filter["types"] = *params.Types
	}
	if params.ShowType != nil {
		filter["show_type"] = *params.ShowType
	}
	// 設置排序選項
	opts := options.Find().SetSkip((params.PageIndex - 1) * params.PageSize).SetLimit(params.PageSize)
	opts.SetSort(bson.D{{Key: "updated_at", Value: -1}})

	// 查詢符合條件的總數
	count, err := repo.Tags.GetClient().CountDocuments(ctx, filter)
	if err != nil {
		return nil, 0, err
	}

	// 執行查詢並獲取結果
	var tags []*entity.Tags
	err = repo.Tags.GetClient().Find(ctx, &tags, filter, opts)
	if err != nil {
		return nil, 0, err
	}

	return tags, count, nil
}

func (repo *TagsRepository) IndexTags20250317001UP(ctx context.Context) (*mongo.Cursor, error) {
	// 等價於 db.account.createIndex({"create_at": 1})
	repo.Tags.PopulateIndex(ctx, "show_type", 1, false)
	repo.Tags.PopulateIndex(ctx, "types", 1, false)
	repo.Tags.PopulateIndex(ctx, "name", 1, false)

	return repo.Tags.GetClient().Indexes().List(ctx)
}
